@@ -16,7 +16,12 @@
 
 #include "aalmediaplayercontrol.h"
 #include "aalmediaplayerservice.h"
-#include "aalmetadatareadercontrol.h"
+#include "aalmediaplaylistcontrol.h"
+#include "aalmediaplaylistprovider.h"
+#include "aalaudiorolecontrol.h"
+#include "aalutility.h"
+
+#include <qmediaplaylistcontrol_p.h>
 
 #include <core/media/service.h>
 #include <core/media/track_list.h>
@@ -24,11 +29,18 @@
 #include <errno.h>
 
 #include <QAbstractVideoSurface>
+#include <QGuiApplication>
 #include <QTimerEvent>
 #include <QThread>
 
 #include <qtubuntu_media_signals.h>
 
+// Uncomment to re-enable media-hub Player session detach/reattach functionality.
+// It is not clear at all whether we should do this or not, as detaching
+// is probably something that should be done when the app finishes, not when it
+// simply moves to the background.
+//#define DO_PLAYER_ATTACH_DETACH
+
 // Defined in aalvideorenderercontrol.h
 #ifdef MEASURE_PERFORMANCE
 #include <QDateTime>
@@ -51,26 +63,99 @@ enum {
 core::Signal<void> the_void;
 }
 
-AalMediaPlayerService::AalMediaPlayerService(QObject *parent):
-    QMediaService(parent),
-    m_hubPlayerSession(NULL),
-    m_playbackStatusChangedConnection(the_void.connect([](){})),
-    m_errorConnection(the_void.connect([](){})),
-    m_mediaPlayerControl(nullptr),
-    m_videoOutput(nullptr),
-    m_metaDataReaderControl(nullptr),
-    m_videoOutputReady(false),
-    m_firstPlayback(true),
-    m_cachedDuration(0),
-    m_mediaPlaylist(NULL)
+AalMediaPlayerService::AalMediaPlayerService(QObject *parent)
+    :
+     QMediaService(parent),
+     m_hubPlayerSession(NULL),
+     m_playbackStatusChangedConnection(the_void.connect([](){})),
+     m_errorConnection(the_void.connect([](){})),
+     m_endOfStreamConnection(the_void.connect([](){})),
+     m_mediaPlayerControl(nullptr),
+     m_videoOutput(nullptr),
+     m_mediaPlaylistControl(nullptr),
+     m_mediaPlaylistProvider(nullptr),
+     m_audioRoleControl(nullptr),
+     m_videoOutputReady(false),
+     m_firstPlayback(true),
+     m_cachedDuration(0),
+     m_mediaPlaylist(NULL),
+     m_doReattachSession(false)
 #ifdef MEASURE_PERFORMANCE
-     , m_lastFrameDecodeStart(0)
-     , m_currentFrameDecodeStart(0)
-     , m_avgCount(0)
-     , m_frameDecodeAvg(0)
+      , m_lastFrameDecodeStart(0)
+      , m_currentFrameDecodeStart(0)
+      , m_avgCount(0)
+      , m_frameDecodeAvg(0)
 #endif
 {
-    m_hubService = media::Service::Client::instance();
+    constructNewPlayerService();
+    // Note: this must be in the constructor and not part of constructNewPlayerService()
+    // or it won't successfully connect to the signal
+    connect(qGuiApp, &QGuiApplication::applicationStateChanged, this, &AalMediaPlayerService::onApplicationStateChanged);
+}
+
+AalMediaPlayerService::AalMediaPlayerService
+    (const std::shared_ptr<core::ubuntu::media::Service> &service, QObject *parent)
+    :
+      QMediaService(parent),
+      m_hubService(service),
+      m_hubPlayerSession(NULL),
+      m_playbackStatusChangedConnection(the_void.connect([](){})),
+      m_errorConnection(the_void.connect([](){})),
+      m_endOfStreamConnection(the_void.connect([](){})),
+      m_mediaPlayerControl(nullptr),
+      m_videoOutput(nullptr),
+      m_mediaPlaylistControl(nullptr),
+      m_mediaPlaylistProvider(nullptr),
+      m_audioRoleControl(nullptr),
+      m_videoOutputReady(false),
+      m_firstPlayback(true),
+      m_cachedDuration(0),
+      m_mediaPlaylist(NULL),
+      m_doReattachSession(false)
+  #ifdef MEASURE_PERFORMANCE
+       , m_lastFrameDecodeStart(0)
+       , m_currentFrameDecodeStart(0)
+       , m_avgCount(0)
+       , m_frameDecodeAvg(0)
+  #endif
+{
+    constructNewPlayerService();
+    // Note: this must be in the constructor and not part of constructNewPlayerService()
+    // or it won't successfully connect to the signal
+    connect(qGuiApp, &QGuiApplication::applicationStateChanged, this, &AalMediaPlayerService::onApplicationStateChanged);
+}
+
+AalMediaPlayerService::~AalMediaPlayerService()
+{
+    m_errorConnection.disconnect();
+    m_playbackStatusChangedConnection.disconnect();
+
+    if (m_audioRoleControl)
+        deleteAudioRoleControl();
+
+    if (m_videoOutput)
+        deleteVideoRendererControl();
+
+    if (m_mediaPlaylistControl)
+        deletePlaylistControl();
+
+    if (m_mediaPlayerControl)
+        deleteMediaPlayerControl();
+
+    if (m_hubPlayerSession)
+        destroyPlayerSession();
+}
+
+void AalMediaPlayerService::constructNewPlayerService()
+{
+    if (not m_hubService.get())
+        m_hubService = media::Service::Client::instance();
+
+    // As core::Connection doesn't allow us to start with a disconnected connection
+    // instance we have to connect it first with a dummy signal and then disconnect
+    // it again. If we don't do this connectSignals() will never be able to attach
+    // to the relevant signals.
+    m_endOfStreamConnection.disconnect();
 
     if (!newMediaPlayer())
         qWarning() << "Failed to create a new media player backend. Video playback will not function." << endl;
@@ -83,7 +168,7 @@ AalMediaPlayerService::AalMediaPlayerService(QObject *parent):
 
     createMediaPlayerControl();
     createVideoRendererControl();
-    createMetaDataReaderControl();
+    createAudioRoleControl();
 
     m_playbackStatusChangedConnection = m_hubPlayerSession->playback_status_changed().connect(
         [this](const media::Player::PlaybackStatus &status) {
@@ -95,18 +180,6 @@ AalMediaPlayerService::AalMediaPlayerService(QObject *parent):
             std::bind(&AalMediaPlayerService::onError, this, _1));
 }
 
-AalMediaPlayerService::~AalMediaPlayerService()
-{
-    m_errorConnection.disconnect();
-    m_playbackStatusChangedConnection.disconnect();
-
-    deleteMediaPlayerControl();
-    if (isVideoSource())
-        deleteVideoRendererControl();
-
-    deleteMetaDataReaderControl();
-}
-
 QMediaControl *AalMediaPlayerService::requestControl(const char *name)
 {
     if (qstrcmp(name, QMediaPlayerControl_iid) == 0)
@@ -125,36 +198,42 @@ QMediaControl *AalMediaPlayerService::requestControl(const char *name)
         return m_videoOutput;
     }
 
-    if (qstrcmp(name, QMetaDataReaderControl_iid) == 0)
+    if (qstrcmp(name, QMediaPlaylistControl_iid) == 0)
     {
-        if (not m_metaDataReaderControl)
-            createMetaDataReaderControl();
+        if (not m_mediaPlaylistControl)
+            createPlaylistControl();
 
-        return m_metaDataReaderControl;
+        // Pass on the media-hub Player object to the playlist control
+        if (m_hubPlayerSession)
+            m_mediaPlaylistControl->setPlayerSession(m_hubPlayerSession);
+
+        return m_mediaPlaylistControl;
     }
 
-    return NULL;
+    if (qstrcmp(name, QAudioRoleControl_iid) == 0)
+    {
+        if (not m_audioRoleControl)
+            createAudioRoleControl();
+
+        return m_audioRoleControl;
+    }
+
+    return nullptr;
 }
 
 void AalMediaPlayerService::releaseControl(QMediaControl *control)
 {
-    if (control == m_mediaPlayerControl)
-        deleteMediaPlayerControl();
-    else if (control == m_videoOutput)
+    if (control == m_videoOutput)
         deleteVideoRendererControl();
-    else if (control == m_metaDataReaderControl)
-        deleteMetaDataReaderControl();
-    else
-        delete control;
 }
 
 bool AalMediaPlayerService::newMediaPlayer()
 {
     // Only one player session needed
-    if (m_hubPlayerSession != NULL)
+    if (m_hubPlayerSession != nullptr)
         return true;
 
-    if (m_hubService == NULL)
+    if (m_hubService == nullptr)
     {
         qWarning() << "Cannot create new media player instance without a valid media-hub service instance";
         return false;
@@ -163,20 +242,30 @@ bool AalMediaPlayerService::newMediaPlayer()
     try {
         m_hubPlayerSession = m_hubService->create_session(media::Player::Client::default_configuration());
     }
-    catch (std::runtime_error &e) {
+    catch (const std::runtime_error &e) {
         qWarning() << "Failed to start a new media-hub player session: " << e.what();
         return false;
     }
 
+    try {
+        // Get the player session UUID so we can suspend/restore our session when the ApplicationState
+        // changes
+        m_sessionUuid = m_hubPlayerSession->uuid();
+    } catch (const std::runtime_error &e) {
+        qWarning() << "Failed to retrieve the current player's uuid: " << e.what() << endl;
+        return false;
+    }
+
     return true;
 }
 
 std::shared_ptr<core::ubuntu::media::video::Sink> AalMediaPlayerService::createVideoSink(uint32_t texture_id)
 {
-    if (not m_hubPlayerSession) throw std::runtime_error
-    {
-        "Cannot create a video sink without a valid media-hub player session"
-    };
+    if (m_hubPlayerSession == NULL)
+        throw std::runtime_error
+        {
+            "Cannot create a video sink without a valid media-hub player session"
+        };
 
     auto sink = m_hubPlayerSession->create_gl_texture_video_sink(texture_id);
     m_videoOutputReady = true;
@@ -187,26 +276,32 @@ std::shared_ptr<core::ubuntu::media::video::Sink> AalMediaPlayerService::createV
 void AalMediaPlayerService::resetVideoSink()
 {
     qDebug() << Q_FUNC_INFO;
+    Q_EMIT SharedSignal::instance()->sinkReset();
     m_firstPlayback = false;
     if (m_videoOutput != NULL)
         m_videoOutput->playbackComplete();
 }
 
-QMediaPlayer::AudioRole AalMediaPlayerService::audioRole() const
+QAudio::Role AalMediaPlayerService::audioRole() const
 {
-    if (m_hubPlayerSession == NULL)
-        return QMediaPlayer::MultimediaRole;
+    if (m_audioRoleControl == nullptr)
+    {
+        qWarning() << "Failed to get audio role, m_audioRoleControl is NULL";
+        return QAudio::UnknownRole;
+    }
 
-    return static_cast<QMediaPlayer::AudioRole>(m_hubPlayerSession->audio_stream_role().get());
+    return m_audioRoleControl->audioRole();
 }
 
-void AalMediaPlayerService::setAudioRole(QMediaPlayer::AudioRole audioRole)
+void AalMediaPlayerService::setAudioRole(QAudio::Role audioRole)
 {
-    if (m_hubPlayerSession == NULL)
+    if (m_audioRoleControl == nullptr)
+    {
+        qWarning() << "Failed to set audio role, m_audioRoleControl is NULL";
         return;
+    }
 
-    qDebug() << __PRETTY_FUNCTION__;
-    m_hubPlayerSession->audio_stream_role().set(static_cast<media::Player::AudioStreamRole>(audioRole));
+    m_audioRoleControl->setAudioRole(audioRole);
 }
 
 void AalMediaPlayerService::setMediaPlaylist(const QMediaPlaylist &playlist)
@@ -227,42 +322,69 @@ void AalMediaPlayerService::setMediaPlaylist(const QMediaPlaylist &playlist)
 
 void AalMediaPlayerService::setMedia(const QUrl &url)
 {
-    if (m_hubPlayerSession == NULL)
+    if (m_hubPlayerSession == nullptr)
     {
         qWarning() << "Cannot open uri without a valid media-hub player session";
         return;
     }
-    if (url.isEmpty())
+
+    // This is critical to allowing a different video source to be able to play correctly
+    // if another video is already playing in the same AalMediaPlayerService instance
+    if (m_videoOutput->textureId() > 0)
     {
-        qWarning() << "Failed to set media source, url must be set." << endl;
-        return;
+        m_mediaPlayerControl->stop();
+        resetVideoSink();
     }
 
     qDebug() << "Setting media to: " << url;
-    const media::Track::UriType uri(url.url().toStdString());
-    try {
-        m_hubPlayerSession->open_uri(uri);
-    }
-    catch (std::runtime_error &e) {
-        qWarning() << "Failed to open media " << url << ": " << e.what();
-        return;
+
+    if (m_mediaPlaylistProvider && url.isEmpty())
+        m_mediaPlaylistProvider->clear();
+
+    if (m_mediaPlaylistProvider == nullptr || m_mediaPlaylistProvider->mediaCount() == 0)
+    {
+        try {
+            m_hubPlayerSession->open_uri(url.toString().toStdString());
+        }
+        catch (const media::Player::Errors::InsufficientAppArmorPermissions &e) {
+            qWarning() << e.what();
+            signalQMediaPlayerError(media::Player::Error::resource_error);
+            return;
+        }
+        catch (const media::Player::Errors::UriNotFound &e) {
+            qWarning() << e.what();
+            signalQMediaPlayerError(media::Player::Error::resource_error);
+            return;
+        }
+        catch (const std::runtime_error &e) {
+            qWarning() << "Failed to set media " << url << ": " << e.what();
+            return;
+        }
     }
 
+    // Make sure this player can be controlled by MPRIS if appropriate
+    updateCurrentPlayer();
+
     if (isVideoSource())
         m_videoOutput->setupSurface();
 }
 
 void AalMediaPlayerService::play()
 {
+    qDebug() << Q_FUNC_INFO;
     if (m_hubPlayerSession == NULL)
     {
         qWarning() << "Cannot start playback without a valid media-hub player session";
         return;
     }
 
-    if (isVideoSource()
-            && !m_firstPlayback && m_videoOutput != NULL)
-        m_videoOutput->setupSurface();
+    if (isVideoSource() && m_videoOutput != NULL)
+    {
+        m_videoOutput->autoPlay(true);
+
+        if (!m_firstPlayback)
+            m_videoOutput->setupSurface();
+    }
 
     // If we previously played and hit the end-of-stream, stop will be called which
     // tears down the video sink. We need a new video sink in order to render video again
@@ -280,11 +402,14 @@ void AalMediaPlayerService::play()
             // Pause all other music and video sessions
             m_hubService->pause_other_sessions(key);
 
+            m_mediaPlayerControl->setMediaStatus(QMediaPlayer::LoadedMedia);
+
             qDebug() << "Actually calling m_hubPlayerSession->play()";
             m_hubPlayerSession->play();
+
             m_mediaPlayerControl->mediaPrepared();
         }
-        catch (std::runtime_error &e) {
+        catch (const std::runtime_error &e) {
             qWarning() << "Failed to start playback: " << e.what();
             return;
         }
@@ -304,7 +429,7 @@ void AalMediaPlayerService::pause()
     try {
         m_hubPlayerSession->pause();
     }
-    catch (std::runtime_error &e) {
+    catch (const std::runtime_error &e) {
         qWarning() << "Failed to pause playback: " << e.what();
         return;
     }
@@ -322,7 +447,7 @@ void AalMediaPlayerService::stop()
         m_hubPlayerSession->stop();
         m_videoOutputReady = false;
     }
-    catch (std::runtime_error &e) {
+    catch (const std::runtime_error &e) {
         qWarning() << "Failed to stop playback: " << e.what();
         return;
     }
@@ -339,7 +464,7 @@ int64_t AalMediaPlayerService::position() const
     try {
         return m_hubPlayerSession->position() / 1e6;
     }
-    catch (std::runtime_error &e) {
+    catch (const std::runtime_error &e) {
         qWarning() << "Failed to get current playback position: " << e.what();
         return 0;
     }
@@ -352,7 +477,12 @@ void AalMediaPlayerService::setPosition(int64_t msec)
         qWarning() << "Cannot set current playback position without a valid media-hub player session";
         return;
     }
-    m_hubPlayerSession->seek_to(std::chrono::microseconds{msec * 1000});
+    try {
+        m_hubPlayerSession->seek_to(std::chrono::microseconds{msec * 1000});
+    }
+    catch (const std::runtime_error &e) {
+        qWarning() << "Failed to set position to " << msec << ": " << e.what();
+    }
 }
 
 int64_t AalMediaPlayerService::duration()
@@ -373,7 +503,7 @@ int64_t AalMediaPlayerService::duration()
         }
         return d / 1e6;
     }
-    catch (std::runtime_error &e) {
+    catch (const std::runtime_error &e) {
         qWarning() << "Failed to get current playback duration: " << e.what();
         return 0;
     }
@@ -390,7 +520,7 @@ bool AalMediaPlayerService::isVideoSource() const
     try {
         return m_hubPlayerSession->is_video_source();
     }
-    catch (std::runtime_error &e) {
+    catch (const std::runtime_error &e) {
         qWarning() << "Failed to check if source is video: " << e.what();
         return false;
     }
@@ -407,7 +537,7 @@ bool AalMediaPlayerService::isAudioSource() const
     try {
         return m_hubPlayerSession->is_audio_source();
     }
-    catch (std::runtime_error &e) {
+    catch (const std::runtime_error &e) {
         qWarning() << "Failed to check if source is video: " << e.what();
         return false;
     }
@@ -424,7 +554,7 @@ int AalMediaPlayerService::getVolume() const
     try {
         return m_hubPlayerSession->volume();
     }
-    catch (std::runtime_error &e) {
+    catch (const std::runtime_error &e) {
         qWarning() << "Failed to get current volume level: " << e.what();
         return 0;
     }
@@ -443,63 +573,97 @@ void AalMediaPlayerService::setVolume(int volume)
 
 void AalMediaPlayerService::createMediaPlayerControl()
 {
-    if (m_hubPlayerSession == NULL)
+    if (m_hubPlayerSession == nullptr)
         return;
 
     m_mediaPlayerControl = new AalMediaPlayerControl(this);
-    m_hubPlayerSession->end_of_stream().connect([this]()
-    {
-        m_firstPlayback = false;
-        Q_EMIT playbackComplete();
-    });
+    connectSignals();
 }
 
 void AalMediaPlayerService::createVideoRendererControl()
 {
-    if (m_hubPlayerSession == NULL)
+    if (m_hubPlayerSession == nullptr)
         return;
 
     m_videoOutput = new AalVideoRendererControl(this);
 }
 
-void AalMediaPlayerService::createMetaDataReaderControl()
+void AalMediaPlayerService::createPlaylistControl()
 {
-    qDebug() << Q_FUNC_INFO;
+    m_mediaPlaylistControl = new AalMediaPlaylistControl(this);
+    m_mediaPlaylistProvider = new AalMediaPlaylistProvider(this);
+    m_mediaPlaylistControl->setPlaylistProvider(m_mediaPlaylistProvider);
+}
 
-    m_metaDataReaderControl = new AalMetaDataReaderControl(this);
+void AalMediaPlayerService::createAudioRoleControl()
+{
+    if (m_hubPlayerSession == nullptr)
+        return;
 
-    if (m_mediaPlayerControl == nullptr)
-        qDebug() << "m_mediaPlayerControl is NULL, can't connect mediaChanged signal";
-
-    connect(m_mediaPlayerControl, SIGNAL(mediaChanged(QMediaContent)),
-            m_metaDataReaderControl, SLOT(onMediaChanged(QMediaContent)));
+    m_audioRoleControl = new AalAudioRoleControl(m_hubPlayerSession);
 }
 
 void AalMediaPlayerService::deleteMediaPlayerControl()
 {
-    if (m_hubPlayerSession == NULL)
+    if (not m_hubPlayerSession)
         return;
 
-    delete m_mediaPlayerControl;
-    m_mediaPlayerControl = NULL;
+    if (m_mediaPlayerControl)
+    {
+        delete m_mediaPlayerControl;
+        m_mediaPlayerControl = nullptr;
+    }
+}
+
+void AalMediaPlayerService::destroyPlayerSession()
+{
+    if (not m_hubPlayerSession)
+        return;
+
+    try {
+        // Invalidates the media-hub player session
+        m_hubService->destroy_session(m_sessionUuid, media::Player::Client::default_configuration());
+        m_sessionUuid.clear();
+
+        // When we arrived here the session is already invalid and we
+        // can safely drop the reference.
+        m_hubPlayerSession = nullptr;
+    }
+    catch (const std::runtime_error &e) {
+        qWarning() << "Failed to destroy existing media-hub player session: " << e.what();
+    }
 }
 
 void AalMediaPlayerService::deleteVideoRendererControl()
 {
-    if (m_hubPlayerSession == nullptr)
-        return;
-
-    delete m_videoOutput;
-    m_videoOutput = nullptr;
+    if (m_videoOutput)
+    {
+        delete m_videoOutput;
+        m_videoOutput = nullptr;
+    }
 }
 
-void AalMediaPlayerService::deleteMetaDataReaderControl()
+void AalMediaPlayerService::deletePlaylistControl()
 {
-    if (m_hubPlayerSession == nullptr)
-        return;
+    if (m_mediaPlaylistProvider)
+    {
+        delete m_mediaPlaylistProvider;
+        m_mediaPlaylistProvider = nullptr;
+    }
+    if (m_mediaPlaylistControl)
+    {
+        delete m_mediaPlaylistControl;
+        m_mediaPlaylistControl = nullptr;
+    }
+}
 
-    delete m_metaDataReaderControl;
-    m_metaDataReaderControl = nullptr;
+void AalMediaPlayerService::deleteAudioRoleControl()
+{
+    if (m_audioRoleControl)
+    {
+        delete m_audioRoleControl;
+        m_audioRoleControl = nullptr;
+    }
 }
 
 void AalMediaPlayerService::signalQMediaPlayerError(const media::Player::Error &error)
@@ -511,10 +675,12 @@ void AalMediaPlayerService::signalQMediaPlayerError(const media::Player::Error &
         case media::Player::Error::resource_error:
             outError = QMediaPlayer::ResourceError;
             outErrorStr = "A media resource couldn't be resolved.";
+            m_mediaPlayerControl->setMediaStatus(QMediaPlayer::InvalidMedia);
             break;
         case media::Player::Error::format_error:
             outError = QMediaPlayer::FormatError;
             outErrorStr = "The media format type is not playable due to a missing codec.";
+            m_mediaPlayerControl->setMediaStatus(QMediaPlayer::InvalidMedia);
             break;
         case media::Player::Error::network_error:
             outError = QMediaPlayer::NetworkError;
@@ -523,6 +689,7 @@ void AalMediaPlayerService::signalQMediaPlayerError(const media::Player::Error &
         case media::Player::Error::access_denied_error:
             outError = QMediaPlayer::AccessDeniedError;
             outErrorStr = "Insufficient privileges to play that media.";
+            m_mediaPlayerControl->setMediaStatus(QMediaPlayer::InvalidMedia);
             break;
         case media::Player::Error::service_missing_error:
             outError = QMediaPlayer::ServiceMissingError;
@@ -548,10 +715,84 @@ void AalMediaPlayerService::onPlaybackStatusChanged()
     switch (m_newStatus)
     {
         case media::Player::PlaybackStatus::ready:
-            break;
         case media::Player::PlaybackStatus::stopped:
-            // FIXME: Disabled for now since this causes next/previous to not work in music-app
-            //m_mediaPlayerControl->setState(QMediaPlayer::StoppedState);
+            m_mediaPlayerControl->setState(QMediaPlayer::StoppedState);
+            break;
+        case media::Player::PlaybackStatus::paused:
+            m_mediaPlayerControl->setState(QMediaPlayer::PausedState);
+            break;
+        case media::Player::PlaybackStatus::playing:
+            // This is necessary in case duration == 0 right after calling play(). At this point,
+            // the pipeline should be 100% prepared and playing.
+            Q_EMIT m_mediaPlayerControl->durationChanged(duration());
+            m_mediaPlayerControl->setState(QMediaPlayer::PlayingState);
+            break;
+        default:
+            qWarning() << "Unknown PlaybackStatus: " << m_newStatus;
+    }
+
+    qDebug() << "PlaybackStatus changed to: " << playbackStatusStr(m_newStatus);
+}
+
+void AalMediaPlayerService::onApplicationStateChanged(Qt::ApplicationState state)
+{
+    try {
+        switch (state)
+        {
+            case Qt::ApplicationSuspended:
+                qDebug() << "** Application has been suspended";
+                break;
+            case Qt::ApplicationHidden:
+                qDebug() << "** Application is now hidden";
+                break;
+            case Qt::ApplicationInactive:
+                qDebug() << "** Application is now inactive";
+#ifdef DO_PLAYER_ATTACH_DETACH
+                disconnectSignals();
+                m_hubService->detach_session(m_sessionUuid, media::Player::Client::default_configuration());
+                m_doReattachSession = true;
+#endif
+                break;
+            case Qt::ApplicationActive:
+                qDebug() << "** Application is now active";
+#ifdef DO_PLAYER_ATTACH_DETACH
+                // Avoid doing this for when the client application first loads as this
+                // will break video playback
+                if (m_doReattachSession)
+                {
+                    m_hubPlayerSession = m_hubService->
+                        reattach_session(m_sessionUuid,
+                                         media::Player::Client::default_configuration());
+                    // Make sure the client's status (position, duraiton, state, etc) are all
+                    // correct when reattaching to the media-hub Player session
+                    updateClientSignals();
+                    connectSignals();
+                }
+#endif
+                updateCurrentPlayer();
+                break;
+            default:
+                qDebug() << "Unknown ApplicationState";
+                break;
+        }
+    } catch (const std::runtime_error &e) {
+        qWarning() << "Failed to respond to ApplicationState change: " << e.what();
+    }
+}
+
+void AalMediaPlayerService::updateClientSignals()
+{
+    qDebug() << Q_FUNC_INFO;
+    if (m_mediaPlayerControl == nullptr)
+        return;
+
+    Q_EMIT m_mediaPlayerControl->durationChanged(duration());
+    Q_EMIT m_mediaPlayerControl->positionChanged(position());
+    switch (m_newStatus)
+    {
+        case media::Player::PlaybackStatus::ready:
+        case media::Player::PlaybackStatus::stopped:
+            m_mediaPlayerControl->setState(QMediaPlayer::StoppedState);
             break;
         case media::Player::PlaybackStatus::paused:
             m_mediaPlayerControl->setState(QMediaPlayer::PausedState);
@@ -564,46 +805,83 @@ void AalMediaPlayerService::onPlaybackStatusChanged()
     }
 }
 
-void AalMediaPlayerService::onError(const core::ubuntu::media::Player::Error &error)
+void AalMediaPlayerService::connectSignals()
+{
+    if (!m_endOfStreamConnection.is_connected())
+    {
+        m_endOfStreamConnection = m_hubPlayerSession->end_of_stream().connect([this]()
+        {
+            m_firstPlayback = false;
+            Q_EMIT playbackComplete();
+        });
+    }
+}
+
+void AalMediaPlayerService::disconnectSignals()
+{
+    if (m_endOfStreamConnection.is_connected())
+        m_endOfStreamConnection.disconnect();
+}
+
+void AalMediaPlayerService::updateCurrentPlayer()
+{
+        // If this player is a multimedia audioRole, then it should possible to
+        // use it for MPRIS control
+    if (audioRole() == QAudio::VideoRole || audioRole() == QAudio::MusicRole)
+    {
+        qDebug() << "Setting player as current player";
+        try {
+            const media::Player::PlayerKey key = m_hubPlayerSession->key();
+            m_hubService->set_current_player(key);
+        } catch (const std::runtime_error &e) {
+            qWarning() << "Failed to set player as current player: " << e.what();
+            return;
+        }
+    }
+}
+
+void AalMediaPlayerService::onError(const media::Player::Error &error)
 {
     qWarning() << "** Media playback error: " << error;
     signalQMediaPlayerError(error);
 }
 
-void AalMediaPlayerService::pushPlaylist()
+QString AalMediaPlayerService::playbackStatusStr(const media::Player::PlaybackStatus &status)
 {
-    if (m_hubPlayerSession == NULL)
+    switch (status)
     {
-        qWarning() << "Cannot push playlist without a valid media-hub player session";
-        return;
-    }
-
-    if (m_mediaPlaylist == NULL)
-        return;
-
-    for (int i = 0; i < m_mediaPlaylist->mediaCount(); i++)
-    {
-        const media::Track::UriType uri(m_mediaPlaylist->media(i).canonicalUrl().url().toStdString());
-        m_hubPlayerSession->track_list()->add_track_with_uri_at(uri, media::TrackList::after_empty_track(), false);
+        case media::Player::PlaybackStatus::ready:
+            return "ready";
+        case media::Player::PlaybackStatus::stopped:
+            return "stopped";
+        case media::Player::PlaybackStatus::paused:
+            return "paused";
+        case media::Player::PlaybackStatus::playing:
+            return "playing";
+        default:
+            qWarning() << "Unknown PlaybackStatus: " << status;
+            return QString();
     }
 }
 
-void AalMediaPlayerService::setPlayer(const std::shared_ptr<core::ubuntu::media::Player> &player)
+void AalMediaPlayerService::setPlayer(const std::shared_ptr<media::Player> &player)
 {
     m_hubPlayerSession = player;
 
     createMediaPlayerControl();
     createVideoRendererControl();
-    createMetaDataReaderControl();
 
-    m_hubPlayerSession->playback_status_changed().connect(
-        [this](const media::Player::PlaybackStatus &status) {
-            m_newStatus = status;
-            QMetaObject::invokeMethod(this, "onPlaybackStatusChanged", Qt::QueuedConnection);
-        });
+    if (!m_playbackStatusChangedConnection.is_connected())
+    {
+        m_playbackStatusChangedConnection = m_hubPlayerSession->playback_status_changed().connect(
+            [this](const media::Player::PlaybackStatus &status) {
+                m_newStatus = status;
+                QMetaObject::invokeMethod(this, "onPlaybackStatusChanged", Qt::QueuedConnection);
+            });
+    }
 }
 
-void AalMediaPlayerService::setService(const std::shared_ptr<core::ubuntu::media::Service> &service)
+void AalMediaPlayerService::setService(const std::shared_ptr<media::Service> &service)
 {
     m_hubService = service;
 }
